@@ -3,18 +3,19 @@
 #include <gmp.h>
 #include "field.h"
 #include "fieldquadratic.h"
+#include "utils.h"
 
 static inline element_ptr fq_nqr(field_ptr f)
 {
-    return ((fq_field_data_ptr) f->data)->nqr;
+    return ((field_ptr) f->data)->nqr;
 }
 
 static void fq_init(element_ptr e)
 {
     fq_data_ptr p = e->data = malloc(sizeof(fq_data_t));
-    fq_field_data_ptr fdp = e->field->data;
-    element_init(p->x, fdp->field);
-    element_init(p->y, fdp->field);
+    field_ptr f = e->field->data;
+    element_init(p->x, f);
+    element_init(p->y, f);
 }
 
 static void fq_clear(element_ptr e)
@@ -265,14 +266,6 @@ static int fq_length_in_bytes(element_ptr e)
     return element_length_in_bytes(p->x) + element_length_in_bytes(p->y);
 }
 
-static void fq_field_clear(field_t f)
-{
-    fq_field_data_ptr p = f->data;
-    mpz_clear(f->order);
-    element_clear(p->nqr);
-    free(p);
-}
-
 static int fq_to_bytes(unsigned char *data, element_t e)
 {
     fq_data_ptr p = e->data;
@@ -293,7 +286,6 @@ static int fq_from_bytes(element_t e, unsigned char *data)
 
 static int fq_is_sqr(element_ptr e)
 {
-    //TODO: this is false:
     //x + y sqrt(nqr) is a square iff x^2 - nqr y^2 is (in the base field)
     fq_data_ptr p = e->data;
     element_t e0, e1;
@@ -350,18 +342,18 @@ static void fq_sqrt(element_ptr n, element_ptr e)
     element_clear(e2);
 }
 
+static void field_clear_fq(field_ptr f)
+{
+    UNUSED_VAR(f);
+    //f->order gets cleared automatically
+}
+
 void field_init_quadratic(field_ptr f, field_ptr fbase)
 {
-    fq_field_data_ptr p;
-
     field_init(f);
-    p = f->data = malloc(sizeof(fq_field_data_t));
 
-    element_init(p->nqr, fbase);
-    //TODO: why bother with p->nqr? even if it's needed,
-    //TODO: why not use element_ptr instead?
-    element_set(p->nqr, field_get_nqr(fbase));
-    p->field = fbase;
+    f->field_clear = field_clear_fq;
+    f->data = fbase;
 
     f->init = fq_init;
     f->clear = fq_clear;
@@ -389,11 +381,9 @@ void field_init_quadratic(field_ptr f, field_ptr fbase)
     f->set1 = fq_set1;
     f->is_sqr = fq_is_sqr;
     f->sqrt = fq_sqrt;
-    f->field_clear = fq_field_clear;
     f->to_bytes = fq_to_bytes;
     f->from_bytes = fq_from_bytes;
 
-    mpz_init(f->order);
     mpz_mul(f->order, fbase->order, fbase->order);
     if (fbase->fixed_length_in_bytes < 0) {
 	f->length_in_bytes = fq_length_in_bytes;
@@ -488,8 +478,21 @@ static void fi_invert(element_ptr n, element_ptr a)
 
 static int fi_is_sqr(element_ptr e)
 {
-    //TODO: this is false:
-    //x + yi is a square iff x^2 + y^2 is (in the base field)
+    //x + yi is a square <=> x^2 + y^2 is (in the base field)
+
+    // Proof: (=>) if x+yi = (a+bi)^2,
+    // then a^2 - b^2 = x, 2ab = y,
+    // thus (a^2 + b^2)^2 = (a^2 - b^2)^2 + (2ab)^2 =  x^2 + y^2
+    // (<=) Suppose A^2 = x^2 + y^2
+    // then if there exist a, b satisfying:
+    //   a^2 = (+-A + x)/2, b^2 = (+-A - x)/2
+    // then (a + bi)^2 = x + yi.
+    // We show that exactly one of (A + x)/2, (-A + x)/2
+    // is a quadratic residue (thus a, b do exist).
+    // Suppose not. Then the product
+    // (x^2 - A^2) / 4 is some quadratic residue, a contradiction
+    // since this would imply x^2 - A^2 = -y^2 is also a quadratic residue,
+    // but we know -1 is not a quadratic residue.
     fq_data_ptr p = e->data;
     element_t e0, e1;
     int result;
@@ -512,8 +515,7 @@ static void fi_sqrt(element_ptr n, element_ptr e)
 
     //if (a+bi)^2 = x+yi then
     //2a^2 = x +- sqrt(x^2 + y^2)
-    //(take the sign which allows a to exist)
-    //and 2ab = y
+    //(take the sign such that a exists) and 2ab = y
     //[thus 2b^2 = - (x -+ sqrt(x^2 + y^2))]
     element_init(e0, p->x->field);
     element_init(e1, e0->field);
@@ -542,12 +544,6 @@ static void fi_sqrt(element_ptr n, element_ptr e)
     element_clear(e2);
 }
 
-static void fi_field_clear(field_t f)
-{
-    free(f->data);
-    //TODO: free f->data->nqr
-}
-
 void element_field_to_fi(element_ptr a, element_ptr b)
 {
     fq_data_ptr p = a->data;
@@ -558,18 +554,22 @@ void element_field_to_fi(element_ptr a, element_ptr b)
 
 static void fi_print_info(FILE *out, field_ptr f)
 {
-    fi_field_data_ptr p = f->data;
+    field_ptr fbase = f->data;
     fprintf(out, "quadratic extension field using sqrt(-1), base field:\n");
-    field_print_info(out, p->field);
+    field_print_info(out, fbase);
+}
+
+static void field_clear_fi(field_ptr f)
+{
+    UNUSED_VAR(f);
+    //f->order gets cleared automatically
 }
 
 void field_init_fi(field_ptr f, field_ptr fbase)
 {
-    fi_field_data_ptr p;
-
     field_init(f);
-    p = f->data = malloc(sizeof(fq_field_data_t));
-    p->field = fbase;
+    f->field_clear = field_clear_fi;
+    f->data = fbase;
     f->init = fq_init;
     f->clear = fq_clear;
     f->set_si = fq_set_si;
@@ -596,12 +596,10 @@ void field_init_fi(field_ptr f, field_ptr fbase)
     f->set1 = fq_set1;
     f->is_sqr = fi_is_sqr;
     f->sqrt = fi_sqrt;
-    f->field_clear = fi_field_clear;
     f->to_bytes = fq_to_bytes;
     f->from_bytes = fq_from_bytes;
     f->print_info = fi_print_info;
 
-    mpz_init(f->order);
     mpz_mul(f->order, fbase->order, fbase->order);
     if (fbase->fixed_length_in_bytes < 0) {
 	f->length_in_bytes = fq_length_in_bytes;
