@@ -625,6 +625,20 @@ static void e_pairing_option_set(pairing_t pairing, char *key, char *value)
     }
 }
 
+void e_pairing_clear(pairing_t pairing)
+{
+    e_pairing_data_ptr p = pairing->data;
+    field_clear(p->Fq);
+    curve_clear(p->Eq);
+    mpz_clear(p->tateexp);
+    free(p);
+
+    mpz_clear(pairing->r);
+    field_clear(pairing->Zr);
+    field_clear(pairing->G1);
+    free(pairing->G1);
+}
+
 void pairing_init_e_param(pairing_t pairing, e_param_t param)
 {
     e_pairing_data_ptr p;
@@ -655,10 +669,11 @@ void pairing_init_e_param(pairing_t pairing, e_param_t param)
     pairing->G1 = malloc(sizeof(field_t));
     pairing->G2 = pairing->G1;
 
-    field_init_curve_group(pairing->G1, p->Eq, param->h);
+    field_init_curve_group(pairing->G1, p->Eq, pairing->r, param->h);
     pairing->GT = p->Fq;
     pairing->phi = phi_identity;
     pairing->option_set = e_pairing_option_set;
+    pairing->clear_func = e_pairing_clear;
 
     element_clear(a);
     element_clear(b);
@@ -829,7 +844,7 @@ void pairing_init_singular_with_node(pairing_t pairing, mpz_t q)
 
     mpz_init(one);
     mpz_set_ui(one, 1);
-    field_init_curve_group(pairing->G1, p->Eq, one);
+    field_init_curve_group(pairing->G1, p->Eq, pairing->r, one);
     mpz_clear(one);
     pairing->GT = p->Fq;
     //pairing->phi = trace;
