@@ -155,6 +155,5 @@ void pairing_init_inp_str(pairing_t pairing, FILE *stream)
 
 void pairing_clear(pairing_t pairing)
 {
-    //TODO: unsafe to call until all pairing types have implemented this
-    if (0) pairing->clear_func(pairing);
+    pairing->clear_func(pairing);
 }
