@@ -17,6 +17,13 @@ struct element_s {
 typedef struct element_s *element_ptr;
 typedef struct element_s element_t[1];
 
+struct element_pp_s {
+    struct field_s *field;
+    void *data;
+};
+typedef struct element_pp_s element_pp_t[1];
+typedef struct element_pp_s *element_pp_ptr;
+
 struct field_s {
     void (*field_clear)(struct field_s *f);
     void (*init)(element_ptr);
@@ -34,6 +41,7 @@ struct field_s {
     void (*mul_si)(element_ptr, element_ptr, signed long int);
     void (*square)(element_ptr, element_ptr);
     void (*doub)(element_ptr, element_ptr); //can't call it "double"!
+    void (*halve)(element_ptr, element_ptr);
     void (*pow_mpz)(element_ptr, element_ptr, mpz_ptr);
     void (*invert)(element_ptr, element_ptr);
     void (*neg)(element_ptr, element_ptr);
@@ -49,10 +57,13 @@ struct field_s {
     int (*from_bytes)(element_ptr, unsigned char *data);
     int (*length_in_bytes)(element_ptr);
     int fixed_length_in_bytes; //length of an element in bytes; -1 for variable
-    mpz_t order; //-1 for infinite order
+    mpz_t order; //0 for infinite order
     element_ptr nqr; //nonquadratic residue
     void (*to_mpz)(mpz_ptr, element_ptr);
     void (*print_info)(FILE *, struct field_s *);
+    void (*pp_init)(element_pp_t p, element_t in);
+    void (*pp_clear)(element_pp_t p);
+    void (*pp_pow)(element_t out, mpz_ptr power, element_pp_t p);
     void *data;
 };
 typedef struct field_s *field_ptr;
@@ -202,6 +213,14 @@ static inline void element_double(element_t n, element_t a)
     n->field->doub(n, a);
 }
 
+/*@manual earith
+Set ''n'' to ''a/2''
+*/
+static inline void element_halve(element_t n, element_t a)
+{
+    n->field->halve(n, a);
+}
+
 /*@manual earith
 Set ''n'' to ''a'' times ''a''.
 */
@@ -223,9 +242,9 @@ static inline void element_to_mpz(mpz_t z, element_t e)
 Generate an element ''e'' deterministically from
 the ''len'' bytes stored in the buffer ''data''.
 */
-static inline void element_from_hash(element_t a, int len, void *data)
+static inline void element_from_hash(element_t e, int len, void *data)
 {
-    a->field->from_hash(a, len, data);
+    e->field->from_hash(e, len, data);
 }
 
 /*@manual epow
@@ -349,18 +368,18 @@ The number of bytes it will write can be determined from calling
 <function>element_length_in_bytes()</function>.
 Returns number of bytes written.
 */
-static inline int element_to_bytes(unsigned char *data, element_t a)
+static inline int element_to_bytes(unsigned char *data, element_t e)
 {
-    return a->field->to_bytes(data, a);
+    return e->field->to_bytes(data, e);
 }
 
 /*@manual etrade
 Reads ''e'' from the buffer ''data'', and returns
 the number of bytes read.
 */
-static inline int element_from_bytes(element_t a, unsigned char *data)
+static inline int element_from_bytes(element_t e, unsigned char *data)
 {
-    return a->field->from_bytes(a, data);
+    return e->field->from_bytes(e, data);
 }
 
 /*@manual epow
@@ -471,4 +490,36 @@ int element_length_in_bytes_compressed(element_t e);
 
 void field_print_info(FILE *out, field_ptr f);
 
+void default_element_pp_init(element_pp_t p, element_t in);
+void default_element_pp_pow(element_t out, mpz_ptr power,
+                              element_pp_t p);
+void default_element_pp_clear(element_pp_t p);
+
+/*@manual epow
+Prepare to exponentiate an element ''in'', and store preprocessing information
+in ''p''.
+*/
+static inline void element_pp_init(element_pp_t p, element_t in) {
+    p->field = in->field;
+    in->field->pp_init(p, in);
+}
+
+/*@manual epow
+Clear ''p''. Should be called after ''p'' is no longer needed.
+*/
+static inline void element_pp_clear(element_pp_t p)
+{
+    p->field->pp_clear(p);
+}
+
+/*@manual epow
+Raise ''in'' to ''power'' and store the result in ''out'', where ''in''
+is a previously preprocessed element, that is, the second argument
+passed to a previous <function>element_pp_init</function> call.
+*/
+static inline void element_pp_pow(element_t out, mpz_ptr power, element_pp_t p)
+{
+    p->field->pp_pow(out, power, p);
+}
+
 #endif //FIELD_H
